/* iksemel (XML parser for Jabber)
** Copyright (C) 2000-2003 Gurer Ozen <madcat@e-kolay.net>
** This code is free software; you can redistribute it and/or
** modify it under the terms of GNU Lesser General Public License.
*/

#include "common.h"
#include "iksemel.h"

struct iks_struct {
	struct iks_struct *next, *prev;
	struct iks_struct *children, *last_child;
	struct iks_struct *attribs, *last_attrib;
	struct iks_struct *parent;
	enum ikstype type;
	ikstack *s;
	char *name;
	char *cdata;
	size_t len;
};

/*****  Node Creating & Deleting  *****/

iks *
iks_new (const char *name)
{
	ikstack *s;
	iks *x;

	s = iks_stack_new (DEFAULT_IKS_CHUNK_SIZE);
	if (!s) return NULL;
	x = iks_new_within (name, s);
	if (!x) {
		iks_stack_delete (s);
		return NULL;
	}
	return x;
}

iks *
iks_new_within (const char *name, ikstack *s)
{
	iks *x;

	x = iks_stack_alloc (s, sizeof (iks));
	if (!x) return NULL;
	memset (x, 0, sizeof (iks));
	x->s = s;
	x->type = IKS_TAG;
	if(name) {
		x->name = iks_stack_strdup (s, name, 0);
		if (!x->name) return NULL;
	}
	return x;
}

iks *
iks_insert (iks *x, const char *name)
{
	iks *y;

	if (!x) return NULL;

	y = iks_new_within (name, x->s);
	if (!y) return NULL;
	y->parent = x;
	if (!x->children) x->children=y;
	if (x->last_child) {
		x->last_child->next = y;
		y->prev = x->last_child;
	}
	x->last_child = y;
	return y;
}

iks *
iks_insert_cdata (iks *x, const char *data, size_t len)
{
	iks *y;

	if(!x || !data) return NULL;
	if(len == 0) len = strlen (data);

	y = x->last_child;
	if (y && y->type == IKS_CDATA) {
		char *tmp;
		tmp = iks_stack_alloc (x->s, y->len + len + 1);
		if (!tmp) return NULL;
		memcpy (tmp, y->cdata, y->len);
		memcpy (tmp + y->len, data, len);
		tmp[y->len + len] = '\0';
		y->cdata = tmp;
		y->len += len;
	} else {
		y = iks_insert (x, NULL);
		if (!y) return NULL;
		y->type = IKS_CDATA;
		y->cdata = iks_stack_alloc (x->s, len + 1);
		if (!y->cdata) return NULL;
		memcpy (y->cdata, data, len);
		y->cdata[len] = '\0';
		y->len = len;
	}
	return y;
}

iks *
iks_insert_attrib (iks *x, const char *name, const char *value)
{
	iks *y;
	size_t len;

	if (!x) return NULL;

	y = x->attribs;
	while (y) {
		if (strcmp (name, y->name) == 0) break;
		y = y->next;
	}
	if (NULL == y) {
		y = iks_new_within (name, x->s);
		if (!y) return NULL;
		y->parent = x;
		if (!x->attribs) x->attribs = y;
		if (x->last_attrib) {
			x->last_attrib->next = y;
			y->prev = x->last_attrib;
		}
		x->last_attrib = y;
	}

	if (value) {
		len = strlen (value);
		y->type = IKS_ATTRIBUTE;
		y->cdata = iks_stack_alloc (x->s, len + 1);
		if (!y->cdata) return NULL;
		memcpy (y->cdata, value, len);
		*(y->cdata + len) = '\0';
		y->len = len;
	} else {
		if (y->next) y->next->prev = y->prev;
		if (y->prev) y->prev->next = y->next;
		if (x->attribs == y) x->attribs = y->next;
		if (x->last_attrib == y) x->last_attrib = y->prev;
	}

	return y;
}

iks *
iks_insert_node (iks *x, iks *y)
{
	y->parent = x;
	if (!x->children) x->children=y;
	if (x->last_child) {
		x->last_child->next = y;
		y->prev = x->last_child;
	}
	x->last_child = y;
	return y;
}

void
iks_hide (iks *x)
{
	iks *y;

	if (!x) return;

	if (x->prev) x->prev->next = x->next;
	if (x->next) x->next->prev = x->prev;
	y = x->parent;
	if (y) {
		if (y->children == x) y->children = x->next;
		if (y->last_child == x) y->last_child = x->prev;
	}
}

void
iks_delete (iks *x)
{
	if (x) iks_stack_delete (x->s);
}

/*****  Node Traversing  *****/

iks *
iks_next (iks *x)
{
	if (x) return x->next;
	return NULL;
}

iks *
iks_next_tag (iks *x)
{
	if (x) {
		while (1) {
			x = x->next;
			if (NULL == x) break;
			if (IKS_TAG == x->type) return x;
		}
	}
	return NULL;
}

iks *
iks_prev (iks *x)
{
	if (x) return x->prev;
	return NULL;
}

iks *
iks_prev_tag (iks *x)
{
	if (x) {
		while (1) {
			x = x->prev;
			if (NULL == x) break;
			if (IKS_TAG == x->type) return x;
		}
	}
	return NULL;
}

iks *
iks_parent (iks *x)
{
	if (x) return x->parent;
	return NULL;
}

iks *
iks_child (iks *x)
{
	if (x) return x->children;
	return NULL;
}

iks *
iks_first_tag (iks *x)
{
	if (x) {
		x = x->children;
		while (x) {
			if (IKS_TAG == x->type) return x;
			x = x->next;
		}
	}
	return NULL;
}

iks *
iks_attrib (iks *x)
{
	if (x) return x->attribs;
	return NULL;
}

iks *
iks_find (iks *x, const char *name)
{
	iks *y;

	if (!x) return NULL;
	y = x->children;
	while (y) {
		if (y->name && strcmp (y->name, name) == 0) return y;
		y = y->next;
	}
	return NULL;
}

char *
iks_find_cdata (iks *x, const char *name)
{
	iks *y;

	y = iks_find (x, name);
	if (!y) return NULL;
	y = y->children;
	if (!y) return NULL;
	return y->cdata;
}

char *
iks_find_attrib (iks *x, const char *name)
{
	iks *y;

	if (!x) return NULL;

	y = x->attribs;
	while (y) {
		if (y->name && strcmp (y->name, name) == 0) return y->cdata;
		y = y->next;
	}
	return NULL;
}

iks *
iks_find_with_attrib (iks *x, const char *tagname, const char *attrname, const char *value)
{
	iks *y;

	if (NULL == x) return NULL;

	if (tagname) {
		for (y = x->children; y; y = y->next) {
			if (IKS_TAG == y->type
				&& strcmp (y->name, tagname) == 0
				&& iks_strcmp (iks_find_attrib (y, attrname), value) == 0) {
					return y;
			}
		}
	} else {
		for (y = x->children; y; y = y->next) {
			if (IKS_TAG == y->type
				&& iks_strcmp (iks_find_attrib (y, attrname), value) == 0) {
					return y;
			}
		}
	}
	return NULL;
}

/*****  Node Information  *****/

ikstack *
iks_stack (iks *x)
{
	if (x) return x->s;
	return NULL;
}

enum ikstype
iks_type (iks *x)
{
	if (x) return x->type;
	return IKS_NONE;
}

char *
iks_name (iks *x)
{
	if (x) return x->name;
	return NULL;
}

char *
iks_cdata (iks *x)
{
	if (x) return x->cdata;
	return NULL;
}

size_t
iks_cdata_size (iks *x)
{
	if (x) return x->len;
	return 0;
}

int
iks_has_children (iks *x)
{
	if (x && x->children) return 1;
	return 0;
}

int
iks_has_attribs (iks *x)
{
	if (x && x->attribs) return 1;
	return 0;
}

/*****  Special  *****/

static void
escape (ikstack *s, char *xml, size_t len)
{
	char c;
	int i;
	int j = 0;

	for (i=0; i<len; i++) {
		c = xml[i];
		if ('&' == c || '<' == c || '>' == c || '\'' == c || '"' == c) {
			if (i - j > 0) iks_stack_strcat (s, xml + j, i - j);
			j = i + 1;
			switch (c) {
			case '&': iks_stack_strcat (s, "&amp;", 5); break;
			case '\'': iks_stack_strcat (s, "&apos;", 6); break;
			case '"': iks_stack_strcat (s, "&quot;", 6); break;
			case '<': iks_stack_strcat (s, "&lt;", 4); break;
			case '>': iks_stack_strcat (s, "&gt;", 4); break;
			}
		}
	}
	if (i - j > 0) iks_stack_strcat (s, xml + j, i - j);
}

char *
iks_string (ikstack *s, iks *x)
{
	int level=0, dir=0;
	iks *y;

	if (!x || x->type != IKS_TAG) return NULL;
	if (!s) return NULL;

	while (1) {
		if (dir==0) {
			if (x->type == IKS_TAG) {
				iks *y;

				iks_stack_strcat (s, "<", 1);
				iks_stack_strcat (s, x->name, 0);
				y = x->attribs;
				while (y) {
					iks_stack_strcatv (s, " ", y->name, "='", s);
					escape (s, y->cdata, y->len);
					iks_stack_strcat (s, "'", 1);
					y = y->next;
				}
				if (x->children) {
					iks_stack_strcat (s, ">", 1);
					x = x->children;
					level++;
					continue;
				} else
					iks_stack_strcat (s, "/>", 2);
			} else
				escape (s, x->cdata, x->len);
		}
		y = x->next;
		if (y) {
			if (0 == level) {
				if (x->children) iks_stack_strcatv (s, "</", x->name, ">", s);
				break;
			}
			x = y;
			dir = 0;
		} else {
			x = x->parent;
			level--;
			if (level >= 0) iks_stack_strcatv (s, "</", x->name, ">", s);
			if (level < 1) break;
			dir = 1;
		}
	}
	return iks_stack_print (s);
}
