/* iksemel (XML parser for Jabber)
** Copyright (C) 2000-2003 Gurer Ozen <madcat@e-kolay.net>
** This code is free software; you can redistribute it and/or
** modify it under the terms of GNU Lesser General Public License.
*/

#include "common.h"
#include "iksemel.h"
#include "perf.h"

#include <sys/stat.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif

#ifdef HAVE_GETOPT_LONG
static struct option longopts[] = {
	{ "stats", 0, 0, 's' },
	{ "timing", 0, 0, 't' },
	{ "memory", 0, 0, 'm' },
	{ "help", 0, 0, 'h' },
	{ "version", 0, 0, 'V' }
};
#endif

static char *shortopts = "stmhV";

static void
print_usage (void)
{
	puts ("Usage: ikslint [OPTIONS] FILE\n"
		"This tool checks the well-formedness of an XML document.\n"
		" -s, --stats      Print statistics.\n"
		"benchmark options:\n"
		" -t, --timing     Print how much time parsing takes.\n"
		" -m, --memory     Print how much memory is used by the parser.\n"
		"help options:\n"
		" -h, --help       Print this text and exit.\n"
		" -V, --version    Print version and exit.\n"
#ifndef HAVE_GETOPT_LONG
		"(long options are not supported on your system)\n"
#endif
		"Report bugs to <iksemel-dev@jabberstudio.org>.");
}

struct {
	iksparser *p;
	unsigned int level;
	/* stats */
	unsigned int max_level;
	unsigned int nr_tags;
	unsigned int nr_stags;
	unsigned int cdata_size;
	unsigned long time;
	int used, arena;
	/* options */
	int pr_stats;
	int pr_time;
	int pr_mem;
} lint;

int
tagHook (void *udata, char *name, char **atts, int type)
{
	switch (type) {
		case IKS_OPEN:
			lint.level++;
			if (lint.level > lint.max_level) lint.max_level = lint.level;
			break;
		case IKS_CLOSE:
			lint.level--;
			lint.nr_tags++;
			break;
		case IKS_SINGLE:
			lint.nr_stags++;
			break;
	}
	return IKS_OK;
}

int
cdataHook (void *udata, char *data, size_t len)
{
	lint.cdata_size += len;
	return IKS_OK;
}

void
lint_reset (void)
{
	if (lint.p) {
		iks_parser_reset (lint.p);
	} else {
		lint.p =iks_sax_new (NULL, tagHook, cdataHook);
		if (NULL == lint.p) exit (2);
	}
	lint.level = 0;
	lint.max_level = 0;
	lint.nr_tags = 0;
	lint.nr_stags = 0;
	lint.cdata_size = 0;
}

void
report_file (void)
{
	if (lint.pr_mem)
		printf ("Memory arena grow: %d bytes, used: %d bytes.\n", lint.arena, lint.used);
	if (lint.pr_time)
		printf ("Parsing took %ld milliseconds.\n", lint.time);
	if (lint.pr_stats) {
		printf ("Tags: %d pairs, %d single, %d max depth.\n", lint.nr_tags, lint.nr_stags, lint.max_level);
		printf ("Total size of character data: %d bytes.\n", lint.cdata_size);
	}
}

void
check_file (char *fname)
{
	FILE *f;
	char *buf;
	struct stat fs;
	size_t sz, blk, ret, pos;
	enum ikserror err;
	int done;

	if (stat (fname, &fs) != 0) {
		fprintf (stderr, "ikslint: Cannot access file '%s'.\n", fname);
		exit (1);
	}
	sz = fs.st_size;
	blk = fs.st_blksize;

	f = fopen (fname, "r");
	if (!f) {
		fprintf (stderr, "ikslint: Cannot open file '%s'.\n", fname);
		exit (1);
	}
	buf = malloc (blk);
	if (!buf) {
		fprintf (stderr, "ikslint: cannot allocate %d bytes.\n", blk);
		exit (2);
	}

	pos = 0;
	done = 0;
	lint.arena = m_arena ();
	lint.used = m_used ();
	t_reset ();
	while (0 == done) {
		ret = fread (buf, 1, blk, f);
		pos += ret;
		if (pos == sz) {
			done = 1;
		} else {
			if (ret != blk) {
				fprintf (stderr, "ikslint: Read error in file '%s'.\n", fname);
				exit (1);
			}
		}
		err = iks_parse (lint.p, buf, ret, done);
		switch (err) {
			case IKS_OK:
				break;
			case IKS_NOMEM:
				exit (2);
			case IKS_BADXML:
				fprintf (stderr, "Invalid xml at byte %ld, line %ld in file '%s'.\n",
					iks_nr_bytes (lint.p), iks_nr_lines (lint.p), fname);
				exit (1);
			case IKS_HOOK:
				exit (2);
		}
	}
	lint.time = t_elapsed ();
	lint.arena = m_arena () - lint.arena;
	lint.used = m_used () - lint.used;

	free (buf);
	fclose (f);
}

int
main (int argc, char *argv[])
{
	int c;

#ifdef HAVE_GETOPT_LONG
	int i;
	while ((c = getopt_long (argc, argv, shortopts, longopts, &i)) != -1) {
#else
	while ((c = getopt (argc, argv, shortopts)) != -1) {
#endif
		switch (c) {
			case 's':
				lint.pr_stats = 1;
				break;
			case 't':
				lint.pr_time = 1;
				break;
			case 'm':
				lint.pr_mem = 1;
				break;
			case 'h':
				print_usage ();
				exit (0);
			case 'V':
				puts ("ikslint (iksemel) "VERSION);
				exit (0);
		}
	}
	for (; optind < argc; optind++) {
		lint_reset ();
		check_file (argv[optind]);
		report_file ();
	}

	return 0;
}
